#include <system/api.h>
#include <system/syscall.h>
#include <string.h>
#include <stdbool.h>
#include <stdint.h>

int errno;
uint16_t freeButtonId = 2;

void kos_WindowRedrawStatus(unsigned int status)
{
	syscall2(12, status);
}

void kos_WindowRedrawStart(void)
{
	kos_WindowRedrawStatus(1);
}

void kos_WindowRedrawFinish(void)
{
	kos_WindowRedrawStatus(2);
}

void kos_DefineAndDrawWindow(unsigned x, unsigned y, unsigned w, unsigned h, unsigned styles, unsigned styles_ex, char *title)
{
	syscall6(0, (x<<16) | w, (y<<16) | h, styles, styles_ex, (unsigned)title);
}

void kos_WriteText(unsigned x, unsigned y, unsigned fncolor, unsigned bgcolor, unsigned style, char* str, unsigned len)
{
	syscall6(4, (x<<16) | y, fncolor | (style&0xF0000000), (unsigned)str, len, bgcolor);
}

int kos_WaitForEvent(void)
{
	return syscall1(10);
}

size_t kos_SetEventMask(size_t mask)
{
	return syscall2(40, mask);
}

int kos_GetKey(unsigned char* ch)
{
	if (ch) {
		int retval = syscall1(2);
		switch (retval & 0xFF) {
			case 0: {
				*ch = (retval >> 8) & 0xFF;
				return 1;
			}
			case 2: {
				*ch = (retval >> 8) & 0xFF;
				return 2;
			}
		}

	}
	return 0;
}

int kos_GetPressedButtonId(void)
{
	uint32_t btn = syscall1(17);
	if (btn == 1)
		return 0;

	uint32_t btnId = btn >> 8;
	return btnId;
}

void kos_GetCursorPosByScreen(t_point *pos)
{
	if (!pos)
		return;
	size_t retval = syscall2(37, 0);
	pos->left = (retval >> 16) & 0xFFFF;
	pos->top  = retval& 0xFFFF;
}

void kos_GetCursorPosByWindow(t_point *pos)
{
	if (!pos)
		return;
	size_t retval = syscall2(37, 1);
	pos->left = (retval >> 16) & 0xFFFF;
	pos->top  = retval& 0xFFFF;
}

size_t kos_GetMouseKeyState(void)
{
	return syscall2(37, 2);
}

void kos_GetSystemTime(t_time *tm)
{
	if (!tm)
		return;
	uint32_t t = syscall1(3);
	tm->hours = t & 0xFF;
	tm->minutes = (t >> 8) & 0xFF;
	tm->seconds = (t >> 16) & 0xFF;
}

int kos_CreateThread(void *proc, void *data)
{
	uint8_t *stack = kos_HeapAlloc(THREAD_STACK_SIZE);
	if (!stack)
		return -1;
	uint32_t *new_esp = (uint32_t *)&stack[THREAD_STACK_SIZE-sizeof(size_t)*2];
	new_esp[0] = (uint32_t)kos_ExitApp;
	new_esp[1] = (uint32_t)data;
	return syscall4(51, 1, (unsigned)proc, (unsigned)new_esp);
}

void kos_ThreadYield(void)
{
	syscall2(68, 1);
}

int kos_GetCurrentThreadId(void)
{
	t_thread_info *info = kos_HeapAlloc(1024);
	if (!info)
		return -1;
	syscall3(9, (unsigned)info, -1);
	size_t tid = info->id;
	kos_HeapFree(info);
	return tid;
}

int kos_GetSlotByTID(int tid)
{
	return syscall3(18, 21, tid);
}

void kos_TerminateThreadBySlot(int slot)
{
	syscall3(18, 2, slot);
}

void kos_TerminateThreadById(int id)
{
	syscall3(18, 18, id);
}

int kos_GetForegroundWindow(void)
{
	return syscall2(18, 7);
}

void kos_SetForegroundWindow(int slot)
{
	syscall3(18, 3, slot);
}

void kos_ExitApp(void)
{
	syscall1(-1);
}

int kos_SendMessage(size_t event, size_t data)
{
	return syscall4(72, 1, event, data);
}

t_thread_info *kos_GetThreadInfo(int thread_slot)
{
	t_thread_info *info = kos_HeapAlloc(1024);
	if (!info)
		return NULL;
	syscall3(9, (unsigned)info, thread_slot);
	return info;
}

void kos_GetWindowSize(int thread_slot, t_rect *rc)
{
	if (!rc)
		return;
	t_thread_info *info = kos_HeapAlloc(1024);
	if (!info)
		return;
	syscall3(9, (unsigned)info, thread_slot);
	rc->left   = info->window_size.left;
	rc->top    = info->window_size.top;
	rc->right  = rc->left + info->window_size.width;
	rc->bottom = rc->top  + info->window_size.height;
	kos_HeapFree(info);
	return;
}

bool kos_Shutdown(int type)
{
	int retval = syscall3(18, 9, type);
	return (retval == 0);
}

void kos_PutPixel(short x, short y, unsigned color)
{
	syscall3(1, MAKEDWORD(x, y), color);
}

void kos_PutLabel(t_label *label)
{
	if (!label)
		return;
	syscall6(
		4, 
		MAKEDWORD(label->position.left, label->position.top), 
		label->font.color | label->font.type,
		(unsigned)label->text,
		0,
		label->font.bgcolor
	);
}

void kos_PutRawImage(void *img, t_rect *rc)
{
	syscall4(7, (unsigned)img, 
		MAKEDWORD(rc->right - rc->left, rc->bottom - rc->top), 
		MAKEDWORD(rc->left, rc->top)
	);
}

void kos_PutButton(t_button *btn)
{
	syscall5(8, 
		MAKEDWORD(btn->size.left, btn->size.right - btn->size.left), 
		MAKEDWORD(btn->size.top, btn->size.bottom - btn->size.top),
		btn->id | btn->type,
		btn->color
	);
}

uint16_t kos_GetFreeButtonId(void)
{
	freeButtonId++;
	return freeButtonId;
}

void kos_DeleteButton(t_button *btn)
{
	syscall5(8, 0, 0, btn->id | BUTTON_DELETE, 0);
}

void kos_Sleep(unsigned usec)
{
	syscall2(5, usec);
}

void kos_GetScreenSize(unsigned *x, unsigned *y)
{
	unsigned retval = syscall1(14);
	if (x) *x = retval >> 16;
	if (y) *y = retval & 0xFFFF;
}

void kos_SystemStyleUpdate(void)
{
	syscall3(48, 0, 0);
}

void kos_SetButtonStyle(int style)
{
	syscall3(48, 1, style);
}

void kos_SetSystemColors(t_system_colors* clrs)
{
	syscall4(48, 2, (unsigned)clrs, sizeof(*clrs));
}

void kos_GetSystemColors(t_system_colors* clrs)
{
	syscall4(48, 3, (unsigned)clrs, sizeof(*clrs));
}

int kos_GetHeightOfSkinHeader(void)
{
	return syscall2(48, 4);
}

void kos_DrawRectangle(unsigned x, unsigned y, unsigned w, unsigned h, unsigned color)
{
	syscall4(13, MAKEDWORD(x,w), MAKEDWORD(y,h), color);
}

void kos_DrawLine(unsigned x, unsigned y, unsigned xto, unsigned yto, unsigned color)
{
	syscall4(38, MAKEDWORD(x,xto), MAKEDWORD(y,yto), color);
}

/*void kos_MouseGetPosByScreen(unsigned *x, unsigned *y)
{
	unsigned retval = syscall2(37, 0);
	if (x) *x = retval >> 16;
	if (y) *y = retval & 0xFFFF;
}

void kos_MouseGetPosByWindow(unsigned *x, unsigned *y)
{
	unsigned retval = syscall2(37, 1);
	if (x) *x = retval >> 16;
	if (y) *y = retval & 0xFFFF;
}*/

t_dll_api * kos_LoadLibrary(char *dllname)
{
	return (t_dll_api *)syscall3(68, 19, (unsigned)dllname);
}

void * kos_GetProcAddress(t_dll_api *fnlist, char *fnname)
{
	if (!fnlist) return 0;
	while (fnlist->name) {
		if (!strcmp(fnlist->name, fnname))
			return fnlist->fn;
		fnlist++;
	}
	return 0;
}

size_t kos_HeapInit(void)
{
	return syscall2(68, 11);
}

void *kos_HeapAlloc(size_t len)
{
	return (void *)syscall3(68, 12, len);
}

int kos_HeapFree(void *mem)
{
	return syscall3(68, 13, (unsigned)mem);
}

void *kos_HeapRealloc(void *mem, size_t len)
{
	return (void *)syscall4(68, 20, len, (unsigned)mem);
}

int kos_ReadFile(char *fname, size_t offset, void *buf, size_t len)
{
	size_t count_rd = 0;
	t_iofile f  = {0};
	f.func      = 0;
	f.offset_lo = offset;
	f.length    = len;
	f.buffer    = buf;
	f.fname0    = 0;
	f.lp_fname  = fname;
	_asm{
		pusha
		mov eax, 70
		lea ebx, f
		call mcall
		mov errno, eax
		mov count_rd, ebx
		popa
	}
	return count_rd;
}

int kos_CreateFile(char *fname, size_t offset, void *buf, size_t len)
{
	size_t count_rw = 0;
	t_iofile f  = {0};
	f.func      = 2;
	f.length    = len;
	f.buffer    = buf;
	f.fname0    = 0;
	f.lp_fname  = fname;
	_asm{
		pusha
		mov eax, 70
		lea ebx, f
		call mcall
		mov errno, eax
		mov count_rw, ebx
		popa
	}
	return count_rw;
}

t_iofile * kos_CreateFileEx(char *fname)
{
	t_iofile *f = kos_HeapAlloc(sizeof(f));
	if (!f)
		return NULL;
	memset(f, 0, sizeof(f));
	
	//f.func      = 2;
	//f.length    = len;
	//f.buffer    = buf;
	f->fname0    = 0;
	f->lp_fname  = fname;
	return f;
}

int kos_ReadFileEx(t_iofile *f, size_t offset, void *buf, size_t len)
{
	if (!f)
		return 0;
	size_t count_rw = 0;
	f->func      = 0;
	f->offset_lo = offset;
	f->length    = len;
	f->buffer    = buf;
	//f->fname0    = 0;
	_asm{
		pusha
		mov eax, 70
		mov ebx, f
		call mcall
		mov errno, eax
		mov count_rw, ebx
		popa
	}
	if (errno == 0 || errno == 6)
		return count_rw;
	return -1;
}

int kos_WriteFile(char *fname, unsigned long long offset, void *buf, size_t len)
{
	size_t count_rw = 0;
	t_iofile f  = {0};
	f.func      = 3;
	f.offset_lo = offset & 0xFFFFFFFF;
	f.offset_hi = offset >> 32;
	f.length    = len;
	f.buffer    = buf;
	f.fname0    = 0;
	f.lp_fname  = fname;
	_asm{
		pusha
		mov eax, 70
		lea ebx, f
		call mcall
		mov errno, eax
		mov count_rw, ebx
		popa
	}
	return count_rw;
}

int kos_TruncateFile(char *fname, unsigned long long size)
{
	t_iofile f  = {0};
	f.func      = 4;
	f.size_lo   = size & 0xFFFFFFFF;
	f.size_hi   = size >> 32;
	f.fname0    = 0;
	f.lp_fname  = fname;
	return syscall2(70, (unsigned)&f);
}

int kos_GetFileInfo(char *fname, t_file_bdfk *fileinfo)
{
	t_iofile f  = {0};
	f.func      = 5;
	f.buffer    = fileinfo;
	f.fname0    = 0;
	f.lp_fname  = fname;
	return syscall2(70, (unsigned)&f);
}

int kos_ExecFile(char *fname, char *arg, unsigned flags)
{
	t_iofile f  = {0};
	f.func      = 7;
	f.flags     = flags;
	f.arg       = arg;
	f.fname0    = 0;
	f.lp_fname  = fname;
	return syscall2(70, (unsigned)&f);
}

int kos_DeleteFile(char *fname)
{
	t_iofile f  = {0};
	f.func      = 8;
	f.fname0    = 0;
	f.lp_fname  = fname;
	return syscall2(70, (unsigned)&f);
}

int kos_MakeDir(char *fname)
{
	t_iofile f  = {0};
	f.func      = 9;
	f.fname0    = 0;
	f.lp_fname  = fname;
	return syscall2(70, (unsigned)&f);
}

void kos_SetCurrentDir(char *dirname)
{
	syscall3(30, 1, (unsigned)dirname);
}

void kos_GetCurrentDir(char *dirname, size_t len)
{
	syscall4(30, 2, (unsigned)dirname, len);
}

t_iofile *kos_FindFisrtFileA(char *path, t_findfileA **find)
{
	if (!path || !find)
		return NULL;
	
	t_iofile *io = kos_HeapAlloc(sizeof(t_iofile));
	if (!io)
		return NULL;
	
	t_findfileA *data = kos_HeapAlloc(sizeof(t_findfileA));
	if (!data)
	{
		kos_HeapFree(io);
		return NULL;
	}
	
	char *delim_pos = strrchr(path, '/');
	if (!delim_pos)
	{
		strcpy(data->find_path, "./");
		strncpy(data->find_mask, path, sizeof(data->find_mask));
	}
	else
	{
		strncpy(data->find_path, path, sizeof(data->find_path));
		data->find_path[delim_pos - path] = 0;
		
		strncpy(data->find_mask, delim_pos+1, sizeof(data->find_mask));
	}
	
	io->index = 0;
	while(1)
	{
		io->func      = 1;
		io->flags2    = 0; // ASCII
		io->length    = 1; // one block
		io->buffer    = &data->header;
		
		io->fname0    = 0;
		io->lp_fname  = data->find_path;
		size_t retval = syscall2(70, (unsigned)io);
		
		if (retval)
		{
			kos_HeapFree(data);
			kos_HeapFree(io);
			return NULL;
		}
		else
		{
			if (String_CompareWithMaskA(data->find_mask, data->bdfk.fname))
			{
				break;
			}
			else
			{
				io->index++;
				continue;
			}
		}
	}
	*find = data;
	
	return io;
}

bool kos_FindNextFileA(t_iofile *io, t_findfileA **find)
{
	if (!io || !find)
		return false;
	
	t_findfileA *data = *find;
	while(1)
	{
		io->index++;
		size_t retval = syscall2(70, (unsigned)io);
		
		if (retval)
		{
			kos_HeapFree(data);
			kos_HeapFree(io);
			return false;
		}
		else
		{
			if (String_CompareWithMaskA(data->find_mask, data->bdfk.fname))
			{
				break;
			}
			else
			{
				continue;
			}
		}
	}
	return true;
}

void kos_IPCInit(void *buffer, size_t size)
{
	if (!buffer)
		return;
	syscall4(60, 1, (unsigned)buffer, size);
}

int kos_IPCSend(int tid, void *buffer, size_t size)
{
	if (!buffer)
		return -1;
	return syscall5(60, 2, tid, (unsigned)buffer, size);
}
